import aiohttp
import asyncio
from collections import OrderedDict
from urllib.parse import urlsplit


from simple_site_crawler.webpage import Webpage


class SiteCrawler:
    """
    Site crawler that uses `asyncio` and `aiohttp` libraries to
    asynchronously crawl the website and create its sitemap
    """
    def __init__(self, url, max_tasks=100):
        """
        Initializes class instance with website URL

        :param url: website URL
        :type url: str
        :param max_tasks: maximum number of asynchronous tasks
        :type max_tasks: int
        """
        url_split = urlsplit(url)
        self.root_url = '{0.scheme}://{0.netloc}/'.format(url_split)
        self.root_domain = '{0.netloc}'.format(url_split)
        self.max_tasks = max_tasks

        self.loop = asyncio.get_event_loop()
        self.queue = asyncio.Queue()

        self.crawled_pages = OrderedDict()

        self.add_page(self.root_url)

    def add_page(self, url):
        """
        Add webpage to the queue if it wasn't crawled before and belongs to
        the root domain

        :param url: webpage url
        :type url: str
        """
        domain = urlsplit(url).netloc

        if url not in self.crawled_pages.keys() and domain == self.root_domain:
            self.queue.put_nowait(url)

    async def crawl_webpage(self, session, url):
        """
        Crawl a webpage

        :param session: aiohttp session instance
        :type session: ClientSession
        :param url: webpage url
        :type url: str
        """
        with aiohttp.Timeout(10):
            async with session.get(url) as response:
                content = await response.text()

        webpage = Webpage(
            url=url,
            content=content,
        )
        self.crawled_pages[url] = webpage

        # Add found links to the queue
        for link in webpage.links:
            self.add_page(link)

    async def worker(self, session):
        """
        Single worker that takes next free URL from the queue and crawls it

        :param session: aiohttp session instance
        :type session: ClientSession
        """
        try:
            while True:
                url = await self.queue.get()
                await self.crawl_webpage(session, url)
                self.queue.task_done()
        except asyncio.CancelledError:
            pass

    async def crawl_website(self):
        """
        Crawl website asynchronously until all found links are crawled
        """
        async with aiohttp.ClientSession() as session:
            workers = [
                asyncio.Task(self.worker(session))
                for _ in range(self.max_tasks)
            ]

            await self.queue.join()

            # Cancel worker after everything is done
            for w in workers:
                w.cancel()



import asyncio
import click
import time

from simple_site_crawler import SiteCrawler
from simple_site_crawler.utils import generate_sitemap_xml, render_children


@click.command()
@click.option('--max-tasks', '-t', type=int, default=100,
              help="Maximum allowed number of async tasks.")
@click.option('--export-to-xml', '-e', is_flag=True, default=False,
              help="Export sitemap to XML file.")
@click.option('--suppress', '-s', is_flag=True, default=False,
              help="Suppress printing output to stdout.")
@click.argument('URL', type=str)
def cli(max_tasks, export_to_xml, suppress, url):
    """
    Simple website crawler that generates its sitemap and can either print it
    (and its static content) or export it to standard XML format.

    See https://github.com/pawelad/simple-site-crawler for more info.
    """
    start_time = time.time()
    loop = asyncio.get_event_loop()

    sitemap = SiteCrawler(
        url=url,
        max_tasks=max_tasks,
    )

    loop.run_until_complete(sitemap.crawl_website())
    end_time = time.time()

    run_time = end_time - start_time

    # Create a standard 'sitemap.xml' from found URLs
    if export_to_xml:
        urls = [page for page in sitemap.crawled_pages.keys()]
        generate_sitemap_xml(urls)

        click.secho(
            "Sitemap saved as 'sitemap.xml'", fg='yellow'
        )
        click.echo()

    # Print sitemap to stdout
    if not suppress:
        for webpage in sitemap.crawled_pages.values():
            click.secho(webpage.url, fg='yellow')

            # We do it this way to nicely print children even if some of the
            # groups are empty and we don't know which will be printed last
            groups = [
                ('Links', webpage.links),
                ('Images', webpage.images),
                ('CSS files', webpage.css),
                ('JavaScript files', webpage.javascript),
            ]
            # Get rid of empty values
            groups = [item for item in groups if item[1]]

            for group_name, group_items in groups[:-1]:
                click.secho('├── {}'.format(group_name), fg='blue')
                click.echo(render_children(group_items))

            last_group_name, last_group_items = groups[-1]
            click.secho('└── {}'.format(last_group_name), fg='blue')
            click.echo(render_children(
                last_group_items,
                prefix='    ├── ',
                last_item_prefix='    └── ',
            ))

            click.echo()

    click.secho(
        "All done! Found and crawled {0} websites in {1:.4f} seconds.".format(
            len(sitemap.crawled_pages), run_time,
        ), fg='green',
    )


if __name__ == '__main__':
    cli()



import os
from tempfile import NamedTemporaryFile

from simple_site_crawler.utils import generate_sitemap_xml, render_children


BASE_DIR = os.path.dirname(os.path.realpath(__file__))


# Tests
def test_generate_sitemap_xml():
    """Test `generate_sitemap_xml()` function"""
    tmp_f = NamedTemporaryFile(delete=False)
    generate_sitemap_xml(
        urls=['http://example.com/'], filename=tmp_f.name,
    )

    # Read generated sitemap.xml content
    with open(tmp_f.name) as f:
        generated_sitemap = f.read()

    # Read expected sitemap.xml content
    file_path = os.path.join(BASE_DIR, 'files', 'sitemap.xml')
    with open(file_path) as f:
        expected_sitemap = f.read()

    assert generated_sitemap == expected_sitemap


def test_render_children():
    """Test `render_children()` function"""
    children = ['child A', 'child B', 'child C']

    expected_outcome = (
        '│   ├── child A\n'
        '│   ├── child B\n'
        '│   └── child C'
    )
    rendered_children = render_children(children)

    assert expected_outcome == rendered_children


def test_render_children_with_custom_prefixes():
    """Test `render_children()` function with custom prefixes"""
    children = ['child A', 'child B', 'child C']

    expected_outcome = (
        '    ├~~child A\n'
        '    ├~~child B\n'
        '    └~~child C'
    )
    rendered_children = render_children(
        children=children,
        prefix='    ├~~',
        last_item_prefix='    └~~',
    )

    assert expected_outcome == rendered_children



import pytest
from click.testing import CliRunner

from simple_site_crawler.command_line import cli


# Fixtures
@pytest.fixture(scope='function')
def runner():
    """Get CliRunner"""
    return CliRunner()


# Tests
# TODO: Mock `SiteCrawler` and don't actually crawl test website
@pytest.mark.parametrize('cli_args', [
    ('http://example.com',),
    ('-t', '5', 'http://example.com'),
    ('--max-tasks', '50', 'http://example.com'),
])
def test_running_the_script(runner, cli_args):
    """Test running the script on 'http://example.com' website"""
    result = runner.invoke(
        cli, args=list(cli_args),
    )

    assert result.exit_code == 0
    assert result.output
    assert result.output.startswith('http://example.com')
    assert 'http://www.iana.org/domains/example' in result.output
    assert result.output.count('\n') == 5


@pytest.mark.parametrize('cli_args', [
    ('-e', 'http://example.com'),
    ('--export-to-xml', 'http://example.com'),
])
def test_exporting_to_xml(runner, mocker, cli_args):
    """Test exporting sitemap to XML file"""
    mocked_generate_sitemap_xml = mocker.patch(
        'simple_site_crawler.command_line.generate_sitemap_xml'
    )

    result = runner.invoke(
        cli, args=list(cli_args),
    )

    assert mocked_generate_sitemap_xml.call_count == 1

    assert result.exit_code == 0
    assert result.output
    assert result.output.count('\n') == 7


def test_suppress_option(runner):
    """Test running the with 'suppress' option"""
    result = runner.invoke(
        cli, args=['-s', 'http://example.com'],
    )

    assert result.exit_code == 0
    assert result.output
    assert result.output.startswith('All done!')
    assert result.output.count('\n') == 1



import os
from uuid import uuid4

import pytest
from bs4 import BeautifulSoup

from simple_site_crawler.webpage import Webpage


BASE_DIR = os.path.dirname(os.path.realpath(__file__))


# Fixtures
@pytest.fixture(scope='module')
def html_content():
    """Return example HTML content"""
    file_path = os.path.join(BASE_DIR, 'files', 'example.html')
    with open(file_path) as f:
        content = f.read()

    return content


# Tests
def test_initialization(html_content):
    """Test class initialization"""
    url = str(uuid4())
    content = html_content

    webpage = Webpage(
        url=url,
        content=content,
    )

    assert webpage.url == url
    assert webpage.content == content

    assert webpage.soup == BeautifulSoup(content, 'html5lib')

    assert webpage.links == webpage.get_website_links()
    assert webpage.images == webpage.get_website_images()
    assert webpage.css == webpage.get_website_css()
    assert webpage.javascript == webpage.get_website_javascript()


def test_get_website_links(html_content):
    """Test `get_website_links()` method"""
    webpage = Webpage(
        url=str(uuid4()),
        content=html_content,
    )

    links = webpage.get_website_links()
    expected_links = [
        'https://google.com/', 'https://www.python.org/',
    ]

    assert set(links) == set(expected_links)


def test_get_website_images(html_content):
    """Test `get_website_images()` method"""
    webpage = Webpage(
        url=str(uuid4()),
        content=html_content,
    )

    images = webpage.get_website_images()
    expected_images = [
        'https://www.google.com/logo.png',
        'https://www.python.org/logo.png'
    ]

    assert set(images) == set(expected_images)


def test_get_website_css(html_content):
    """Test `get_website_css()` method"""
    webpage = Webpage(
        url=str(uuid4()),
        content=html_content,
    )

    css = webpage.get_website_css()
    expected_css = [
        'https://fonts.googleapis.com/css?family=Roboto',
        'https://code.getmdl.io/1.3.0/material.indigo-pink.min.css'
    ]

    assert set(css) == set(expected_css)


def test_get_website_javascript(html_content):
    """Test `get_website_javascript()` method"""
    webpage = Webpage(
        url=str(uuid4()),
        content=html_content,
    )

    js = webpage.get_website_javascript()
    expected_js = [
        'https://ajax.googleapis.com/ajax/libs/jquery/3.1.1/jquery.min.js',
        'https://code.getmdl.io/1.3.0/material.min.js'
    ]

    assert set(js) == set(expected_js)



import xml.etree.cElementTree as ET
from xml.dom import minidom


def generate_sitemap_xml(urls, filename='sitemap.xml'):
    """
    Generate 'sitemap.xml' file based on passed urls

    :param urls: list of URLs
    :type urls: list of str
    :param filename: output file name
    :type filename: str
    """
    urlset = ET.Element(
        'urlset',
        xmlns='http://www.sitemaps.org/schemas/sitemap/0.9',
    )

    for webpage_url in urls:
        url = ET.SubElement(urlset, 'url')
        ET.SubElement(url, 'loc').text = webpage_url

    pretty_xml = minidom.parseString(
        ET.tostring(urlset),
    ).toprettyxml(
        encoding='UTF-8',
    )

    with open(filename, 'wb') as f:
        f.write(pretty_xml)


def render_children(children, prefix='│   ├── ', last_item_prefix='│   └── '):
    """
    Simple helper for rendering element children

    :param children: children container
    :type children: iterable object
    :param prefix: string that will prefix the child element
    :type prefix: str
    :param last_item_prefix: string that will prefix the last child elemment
    :return: rendered children string
    :rtype: str
    """
    s = ''

    for item in children[:-1]:
        s += '{}{}\n'.format(prefix, item)

    s += '{}{}'.format(last_item_prefix, children[-1])

    return s



import asyncio
import random
from collections import OrderedDict

from simple_site_crawler.site_crawler import SiteCrawler


# Tests
def test_initialization(mocker):
    """Test class initialization"""
    mocked_add_page = mocker.patch(
        'simple_site_crawler.site_crawler.SiteCrawler.add_page',
    )

    url = 'http://example.co.uk/subpage'
    max_tasks = random.randint(0, 100)

    site_crawler = SiteCrawler(
        url=url,
        max_tasks=max_tasks,
    )

    assert site_crawler.root_url == 'http://example.co.uk/'
    assert site_crawler.root_domain == 'example.co.uk'
    assert site_crawler.max_tasks == max_tasks

    assert isinstance(site_crawler.loop, type(asyncio.get_event_loop()))
    assert isinstance(site_crawler.queue, asyncio.Queue)
    assert isinstance(site_crawler.crawled_pages, OrderedDict)

    mocked_add_page.assert_called_once_with(site_crawler.root_url)


def test_add_page(mocker):
    """Test `add_page()` method"""
    mocked_add_page = mocker.patch(
        'asyncio.Queue.put_nowait',
    )

    site_crawler = SiteCrawler('http://example.com/')
    site_crawler.crawled_pages.update({
        'http://example.com/crawled_page': None,
    })

    site_crawler.add_page('https://google.com/')
    site_crawler.add_page('http://example.com/crawled_page')
    site_crawler.add_page('http://example.com/subpage')

    mocked_add_page.assert_called_with('http://example.com/subpage')
    assert mocked_add_page.call_count == 2


# TODO: Add coroutines tests
